--------------------------------------------------------------------------------
--
--  Murphi Model of OCap patterns: Membranes, forwarders and revocation
--  Deian Stefan
--
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- | Constants, types and variables
--------------------------------------------------------------------------------


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
const
  EXAMPLE_CASE: 4;
  NumObjects:   5;     -- number of objects
  NetworkSize:  2;     -- max. number of outstanding messages in network
   -- set = 1 in PL case
   --     > 1 in OS case
  LIMIT_GRANT: true;   -- Don't give capability to self or a revoker
  MAX_NR_MESSAGES: 100;

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
type
  ObjectId:   1..NumObjects;
  TimeStamp:  0..MAX_NR_MESSAGES;

  
  Message : enum {   -- [] denotes optional, {} denotes implicit
    M_Call,          -- Call(return-capability)
    M_Return,        -- Return([isRevoked])
    M_Grant,         -- Grant({return-capability}, capability)
    M_Revoke         -- Revoke(return-capability)
  };

  MessageArgs : record
    cap:       ObjectId; -- Grant/return capability
    isRevoked: boolean;  -- Forwarder is revoked
  end;
  -- | Inter Object Communication message
  IOCMessage : record    
    source:    ObjectId;    -- Souce of message
    dest:      ObjectId;    -- Destination of message
    message:   Message;     -- Message
    args:      MessageArgs; -- Message arguments
    isWrapped: boolean;     -- Is the message wrapped
    onPath:    array[ObjectId] of boolean; -- Objects on path from source->dest
    tstamp:     TimeStamp;  -- Time stamp
  end;

  -- | Object state
  ObjectState : enum {
    O_IDLE,     -- Object is in idle state
    O_BLOCK     -- Object waiting for receiver to return message
  };

  -- | Cabaility type
  CapType : enum {
    CAP_NULL,           -- No capability
    CAP_RAW,            -- Unwrapped, i.e., raw capability,
    CAP_WRAPPED         -- Wrapped capability
  };

  -- | State of a revocable forwarder
  RevocableForwarderState : record 
     Source:      ObjectId; -- Current forwarding source
     Dest:        ObjectId; -- Forwarding destination
     isRevocable: boolean;  -- Is Revocable forwarder
     Revoker:     ObjectId; -- The revoker to check against before forwarding
     iocMessage:  IOCMessage; -- Message saved by revocable forwarder
  end;

  RevokerState : record
    enabled:  boolean;    -- Boolean object controling access
    changedAt: TimeStamp; -- Time boolean was toggled
  end;

  -- | Objects.
  -- Forwarder:
  --  A forwarding object simply forwards messages to the fwdDestination
  --  object. If the object is not also a membrane, then it forwards the
  --  message as received.
  -- Membrane:
  --  A membrane object wraps all its messages. Although it is common for a 
  --  membrane object to be a forwarder, this is not an imposed requirement.
  -- 
  Object : record
    state:       ObjectState;                -- Current object state
    cList:       array[ObjectId] of CapType; -- List of capabilites
    taintList:   array[ObjectId] of boolean; -- Object it raw communicated with

    -- Membrane objects:
    isMembrane:  boolean;                    -- Object is a membrane
    -- Forwarding objects:
    isForwarder: boolean;                   -- Object is a forwarder
     fwdST:      RevocableForwarderState;   --  with state fwdST
    -- Revokable objects:
    isRevoker:   boolean;                   -- Object is a revoker
     revST:      RevokerState;              --  with state revST
    -- Limited forwarding/revokable objects:
    isLimited:   boolean;
  end;


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var                                            -- state variables for
  net: multiset[NetworkSize] of IOCMessage;    --  network
  obj: array[ObjectId] of Object;              --  objects
  curTime: TimeStamp;                          --  time

--------------------------------------------------------------------------------
-- | Helper functions/procedured
--------------------------------------------------------------------------------
  procedure incTime ();
  begin
    curTime := curTime+1;
  end;

  procedure addMessageToNet ( var  m : IOCMessage );
  begin
    m.tstamp:=curTime;
    incTime();
    multisetadd(m,net);
  end;
  -- | Clear the path, setting only the current object i
  procedure resetPath ( i : ObjectId ; var outM : IOCMessage );
  begin
    for l : ObjectId do
      outM.onPath[l] := false;
    end;
    outM.onPath[i] := true;
  end;

  -- | If O_i received a message that is not wrapped then
  -- the paths from all objects on the path to O_i are tainted
  procedure updateTaintList( i : ObjectId ; inM : IOCMessage );
  begin
    if !inM.isWrapped then
      for k : ObjectId do
        if inM.onPath[k] then
          obj[i].taintList[k] := true; -- path from O_k to O_i is tainted
        end
      end;
    end;
  end;

  -- | Forwarder O_i received a message, forward it. If the message
  -- is a return then send it to the original sender.
  procedure doForwardMessage( i : ObjectId ; inM : IOCMessage );
    var outM : IOCMessage;
  begin
    updateTaintList(i,inM);

    undefine outM;
    outM.source       := i;
    outM.message      := inM.message;
    outM.onPath       := inM.onPath;
    outM.onPath[i]    := true;

    if inM.message != M_Return then
      outM.dest           := obj[i].fwdST.Dest;
      outM.args           := inM.args;
      outM.isWrapped      := inM.isWrapped | obj[i].isMembrane
                             | (obj[i].cList[obj[i].fwdST.Dest] = CAP_WRAPPED);
      obj[i].fwdST.Source := inM.source;
      obj[i].state        := O_BLOCK;
    else
      outM.isWrapped    := inM.isWrapped | obj[i].isMembrane;
      outM.dest         := obj[i].fwdST.Source;
      obj[i].state      := O_IDLE;
    end;
    addMessageToNet (outM);

  end;

  procedure doForwardFailMessage( i : ObjectId ; inM : IOCMessage );
    var outM : IOCMessage;
  begin
    -- Capability has been revoked, reply back to source with fail
    undefine outM;
    outM.source         := i;
    outM.dest           := obj[i].fwdST.Source;
    outM.message        := M_Return;
    outM.args.isRevoked := true;
    outM.onPath         := inM.onPath;
    outM.onPath[i]      := true;
    outM.isWrapped      := obj[i].isMembrane | inM.isWrapped;
    obj[i].state        := O_IDLE;
    addMessageToNet (outM);
  end; 

--------------------------------------------------------------------------------
-- | Example start states
--------------------------------------------------------------------------------
  procedure doExample1();
  begin
  -- | Description:
  -- Attenuated forwarder pattern.
  -- O_4 is an attenuated forwarder to O_3.
  --                  O_2       
  --                   ^ 
  --                   |
  --                   |
  --                   |
  --                   |
  --                   o 
  --                  O_1o---------.->O_3
  --                   o          /
  --                   |         /
  --                   |        /
  --                   |       / 
  --                   |      /
  --                   v     /
  --                  O_4o--'

    assert (NumObjects >= 4) "Example requires at least 4 objects";

    for j : ObjectId do
      obj[1].cList[j] := CAP_RAW;       -- O_1 object has capability to all
    end;

    obj[4].isForwarder        := true;         -- O_4 is a Forwarder
    obj[4].isMembrane         := true;         -- O_4 is a Membrane
    obj[4].cList[3]           := CAP_WRAPPED;  -- to to O_3
    obj[4].fwdST.Dest         := 3;            -- to to O_3
    obj[4].fwdST.isRevocable  := false; 
  end;


  procedure doExample2();
  begin
  -- | Description:
  -- O_4 is a membrane object with capability to O_2 being wrapped.
  -- O2 is a non-attenuated forwarder to O_3.
  --         
  --                  O_1o---------.->O_2o--------->O_3
  --                   o          /
  --                   |         /
  --                   |        /
  --                   |       / 
  --                   |      /
  --                   v     /
  --                  O_4o--'

    assert (NumObjects >= 4) "Example requires at least 4 objects";

    obj[1].cList[2] := CAP_RAW;         -- O_1 object has capability O_2

    obj[4].isMembrane   := true;        -- O_4 is a Membrane, with cap to O_2
    obj[4].cList[2]     := CAP_WRAPPED;

    obj[2].isForwarder        := true;   -- O_2 is a Forwarder to O_3
    obj[2].cList[3]           := CAP_RAW;
    obj[2].fwdST.Dest         := 3;
    obj[2].fwdST.isRevocable  := false; 
  end;

  procedure doExample3();
  begin
  -- | Description:
  -- O_4 is an attenuated (revokable) forwarder to O_5, with object O_3
  -- being the revoker.
  -- 
  --                  O_2       
  --                   ^ 
  --                   |
  --                   |
  --                   |
  --                   |
  --                   o
  --                  O_1 o-------.
  --                 o   o         \
  --                /     \          \
  --               /       \          \
  --              |         |          |
  --              v         v          v
  --          R(O_3)<----oF(O_4)o---->O_5

    assert (NumObjects >= 5) "Example requires at least 5 objects";

    for j : ObjectId do
      obj[1].cList[j] := CAP_RAW;       -- O_1 object has capability to all
    end;
                                         
    obj[3].isRevoker       := true;        -- O_3 is a Revoker
    obj[3].revST.enabled   := true;        --  not revoked
    obj[3].revST.changedAt := curTime;     --  time it was changed

    obj[4].isForwarder        := true;   -- O_4 is a Forwarder to O_5
    obj[4].isMembrane         := true;   -- O_4 is a Membrane 
    obj[4].cList[5]           := CAP_RAW;
    obj[4].fwdST.Dest         := 5;
    obj[4].fwdST.isRevocable  := true; 
    obj[4].fwdST.Revoker      := 3; 
  end;

  procedure doExample4();
  begin
  -- | Description:
  -- O_4 is an attenuated (limited) revokable forwarder to O_5, with object O_3
  -- being the limited forwarder the owner, O_1, grants to others.
  -- 
  --                  O_2       
  --                   ^ 
  --                   |
  --                   |
  --                   |
  --                   |
  --                   o
  --                  O_1 o-------.
  --                 o   o         \
  --                /     \          \
  --               /       \          \
  --              |         |          |
  --              v         v          v
  --          F(O_3)o---->R(O_4)o---->O_5
  -- F = limited forwarder
  -- R = revokerk/forwarder

    assert (NumObjects >= 5) "Example requires at least 5 objects";

    for j : ObjectId do
      obj[1].cList[j] := CAP_RAW;       -- O_1 object has capability to all
    end;
                                         
    obj[3].isMembrane    := true;   -- O_3 is a membrane 
    obj[3].isForwarder   := true;   -- O_3 is a forwarder
    obj[3].isLimited     := true;   -- O_3 is limited
    obj[3].fwdST.Dest    := 4;      -- O_3 is a forwarder to O_4
    obj[3].cList[4]      := CAP_WRAPPED;
    obj[3].fwdST.isRevocable  := false;  -- O_4 is not revocable

    obj[4].isLimited          := true;   -- O_4 is "limited"
    obj[4].isForwarder        := true;   -- O_4 is a forwarder
    obj[4].isRevoker          := true;   -- O_4 is a revoker
    obj[4].revST.enabled      := true;   --     not revoked
    obj[4].revST.changedAt    := curTime;--     time it was changed
    obj[4].cList[5]           := CAP_RAW;
    obj[4].fwdST.Dest         := 5;
    obj[4].fwdST.isRevocable  := false;  -- O_4 is not revocable
    obj[4].fwdST.Revoker      := 1;      -- O_4 can be revoked by O_1
  end;


--------------------------------------------------------------------------------
-- | Rules
--------------------------------------------------------------------------------

-- | Non-forwarding object O_i sends Call(O_i) or Revoke(O_i) message to O_j,
-- i!=j. Message includes  capability to O_i as to let the receiver send a
-- Return message. If the capability to O_j is wrapped, or if O_i is a
-- memberane object, then the message is wrapped.
-- If O_j is a Revoker send Revoke, else send Call.
-- 
-- Once the message is sent, O_i blocks, waiting for return.
--
ruleset i: ObjectId do
  ruleset j: ObjectId do
    rule "Simple object sends Call/Revoke message"
      
      i != j                      &     -- O_i is not O_j
      obj[i].state = O_IDLE       &     -- O_i in idle state
      !obj[i].isForwarder         &     -- O_i is not a forwarder
      !obj[i].isRevoker           &     -- O_i is not a revoker
      !obj[i].isLimited           &     -- O_i is not limited
      obj[i].cList[j] != CAP_NULL &     -- O_i has capability to O_j 

      multisetcount (l:net, true) < NetworkSize
      
    ==>

    var
      outM: IOCMessage;

    begin
      undefine outM;
      outM.source       := i;
      outM.dest         := j;
      outM.message      := obj[j].isRevoker? M_Revoke : M_Call;
      outM.args.cap     := i;   -- capability used to return
      outM.isWrapped    := (obj[i].cList[j] = CAP_WRAPPED) | obj[i].isMembrane;

      resetPath(i,outM);

      obj[i].state      := O_BLOCK;
      addMessageToNet (outM);

    end;
  end;
end;


-- | Simple object O_i sends Grant(O_k) message to O_j, i!=j, j!=k
-- Grant message to O_j includes capability to O_k. If O_i's capability to O_k
-- is wrapped, if O_i is a membrane object, or if O_i's capability to O_j is
-- wrapped, then the message is wrapped, i.e., O_j's capability to  O_k will be
-- wrapped.
-- 
-- Note that O_i never grants cabpability to a limited revoker/forwarder.
-- Once the message is sent, O_i blocks, waiting for return.
--
ruleset i: ObjectId do
  ruleset j: ObjectId do
    ruleset k: ObjectId do
      rule "Simple object sends Grant message"
        
        i != j & j != k                &     -- O_i != O_j and O_j != O_k
        obj[i].state = O_IDLE          &     -- O_i in idle state
        !obj[i].isForwarder            &     -- O_i is not a forwarder
        !obj[i].isRevoker              &     -- O_i is not a revoker
        !obj[i].isLimited              &     -- O_i is not limited
        obj[i].cList[j] != CAP_NULL    &     -- O_i has capability to O_j 
        obj[i].cList[k] != CAP_NULL    &     -- O_i has capability to O_k 
        !(obj[k].isRevoker &
                     obj[k].isLimited) &     -- O_k is not limited revoker/fwd
        (!LIMIT_GRANT  |                     -- if limit then O_i != O_j
         (i != k & !obj[k].isRevoker)) &     -- O_k is not a revoker

        multisetcount (l:net, true) < NetworkSize
        
      ==>

      var
        outM: IOCMessage;

      begin
        undefine outM;
        outM.source       := i;
        outM.dest         := j;
        outM.message      := M_Grant;
        outM.args.cap     := k;
        outM.isWrapped    := obj[i].isMembrane 
                             | (obj[i].cList[j] = CAP_WRAPPED)
                             | (obj[i].cList[k] = CAP_WRAPPED);

        resetPath(i,outM);

        obj[i].state      := O_BLOCK;
        addMessageToNet (outM);

      end;
    end;
  end;
end;

-- | Simple (not forwarder/revoker) object O_i receives message.
-- If the message is a Call(O_j) or Grant(O_k) then O_i sends a Return()
-- message to O_j. If the message is wrapped, or O_i is a membrane object, then
-- the return message is also wrapped.
--
-- If the message is a Return() then O_i goes back to idle.
--
-- If the message is a Grant(O_k) and O_i does not already have a capability
-- to O_i, O_k is added to O_i's capability list. If the message is wrapped
-- then the capability is saved as a wrapped capability and not raw.
--
ruleset i: ObjectId do
  choose j: net do
    rule "Simple object receives message"
      
      net[j].dest = i            &      -- Message for O_i
      !obj[i].isForwarder        &      -- O_i in not a forwarder
      !obj[i].isRevoker          &      -- O_i in not a revoker
      !obj[i].isLimited          &      -- O_i is not limited
      net[j].message != M_Revoke &      -- Should not be Revoke()
      ((obj[i].state = O_IDLE & net[j].message != M_Return) |
       (obj[i].state = O_BLOCK& net[j].message = M_Return) ) 

    ==>

    var
      inM:  IOCMessage;
      outM: IOCMessage;

    begin
      inM := net[j];
      multisetremove (j,net);

      updateTaintList(i,inM);

--      assert (inM.message != M_Revoke) "Non-revoker received Revoke()";

      if inM.message = M_Return then
        obj[i].state      := O_IDLE;
      else


        undefine outM;
        outM.source       := i;
        outM.dest         := inM.source;
        outM.message      := M_Return;
        outM.isWrapped    := inM.isWrapped | obj[i].isMembrane;

        resetPath(i,outM);

        if inM.message = M_Grant & obj[i].cList[inM.args.cap] = CAP_NULL then
          obj[i].cList[inM.args.cap] := outM.isWrapped ? CAP_WRAPPED : CAP_RAW;
        end;

        addMessageToNet (outM); -- send Return()
      end;

    end;
  end;
end;

-- | Revoker object O_i receives message.
-- If the message is a Revoke(O_j) then O_i toggles state and sends a Return()
-- message to O_j. If the message is wrapped the return message is also wrapped.
--
-- If the message is a Call(O_j)  then O_j sends O_j a return message with
-- the isRevoked arugment set to the status of it's state.
--
ruleset i: ObjectId do
  choose j: net do
    rule "Revoker object receives message"
      
      net[j].dest = i           &      -- Message for O_i
      obj[i].isRevoker          &      -- O_i is a revoker
      !obj[i].isLimited         &      -- O_i is not limited
      (net[j].message = M_Revoke|      -- Message is Revoke()
       net[j].message = M_Call) &      --  or Call()
      ((obj[i].state = O_IDLE & net[j].message != M_Return) |
       (obj[i].state = O_BLOCK& net[j].message = M_Return) )

    ==>

    var
      inM:  IOCMessage;
      outM: IOCMessage;

    begin
      inM := net[j];
      multisetremove (j,net);

      updateTaintList(i,inM);

--      assert (inM.message = M_Revoke | inM.message = M_Call )
--           "Revoker received message other than Revoke() and Call()";

      undefine outM;
      outM.source       := i;
      outM.dest         := inM.source;
      outM.message      := M_Return;
      outM.isWrapped    := inM.isWrapped | obj[i].isMembrane;

      resetPath(i,outM);

      if inM.message = M_Revoke then
        -- toggle:
        obj[i].revST.enabled := !obj[i].revST.enabled;
        obj[i].revST.changedAt := curTime;
      end;

      outM.args.isRevoked := !obj[i].revST.enabled;

      addMessageToNet (outM); -- send Return([isRevoked])

    end;
  end;
end;

-- | Non-revocable forwarder O_i receives and forwards message.
-- If O_i is a forwarder for O_j and receives a message from O_k, then
-- O_i forwards the message to O_j.
--
-- If the message is wrapped, O_i is a membrane, or the capability
-- O_i has to O_j is wrapped then the forwarded message is also wrapped.
-- Additionally, O_i adds itself to the message path list.
--
-- On Return the result is wrapped only if the message is wrapped,
-- or if O_i is a membrane. Note that if the first non-Return message
-- that O_i forwarded was wrapped (because the capability to O_j was 
-- wrapped) the return will be wrapped as well.
--
-- If O_i is a limited forwarder, then it does not forward Revoke()
--
-- Once the non-Return message is sent, O_i blocks, waiting for Return.
--
-- Receving a Return message, O_i goes back into idle state, forwarding
-- the Return to the original invoker.
--
ruleset i: ObjectId do
  choose j: net do
    rule "Non-revocable (limited) forwarding object receives & forwards message"
      
      net[j].dest = i                        & -- Message for O_i
      obj[i].isForwarder                     & -- O_i is a forwarder
      !obj[i].isRevoker                      & -- O_i is not a revoker
      !obj[i].fwdST.isRevocable              & -- O_i is not revocable
      (obj[i].isLimited &                      -- O_i is a limited 
        net[j].message != M_Revoke)          & --   with no Revoke()
      ((obj[i].state = O_IDLE & net[j].message != M_Return) |
       (obj[i].state = O_BLOCK& net[j].message = M_Return) )

    ==>

    var
      inM:  IOCMessage;

    begin
      inM := net[j];
      multisetremove (j,net);

      doForwardMessage(i,inM);

    end;
  end;
end;


-- | Revocable forwarder O_i receives and forwards message.
-- If O_i receives a Return(isRevoked) from it's Revoker
-- then if it has been Revoked it sends a (fail) Return to the original source
-- of the message and goes back into idle state.
-- On the other hand, if the capability has not been revoked, it
-- forwards the message, the same as a non-revocable forwarding object.
-- 
-- If O_i receives a message other than Return(isRevoked) from it's Revoker
-- it sends a Call message to it's Revoker and blocks, waiting for a Return
-- indicating whether it should forward or not. Note that before blocking
-- O_i saves the message it received (in fwdST.iocMessage).

ruleset i: ObjectId do
  choose j: net do
    rule "Revocable forwarding object receives and forwards message"
      
      net[j].dest = i                       & -- Message for O_i
      obj[i].isForwarder                    & -- O_i is a forwarder
      !obj[i].isLimited                     & -- O_i is not limited
      obj[i].fwdST.isRevocable              & -- O_i is revocable
      ((obj[i].state = O_IDLE & net[j].message != M_Return) |
       (obj[i].state = O_BLOCK& net[j].message = M_Return) )

    ==>

    var
      inM:  IOCMessage;
      outM: IOCMessage;

    begin
      inM := net[j];
      multisetremove (j,net);

      if inM.message = M_Return & inM.source = obj[i].fwdST.Revoker then
          if inM.args.isRevoked then
            -- Capability has been revoked, reply back to source with fail
            doForwardFailMessage(i, obj[i].fwdST.iocMessage);
        /*
            undefine outM;
            outM.source         := i;
            outM.dest           := obj[i].fwdST.Source;
            outM.message        := M_Return;
            outM.args.isRevoked := true;
            outM.onPath         := obj[i].fwdST.iocMessage.onPath;
            outM.onPath[i]      := true;
            outM.isWrapped      := obj[i].isMembrane 
                                   | obj[i].fwdST.iocMessage.isWrapped;
            obj[i].state        := O_IDLE;
            addMessageToNet (outM);
            */
          else
            -- Still have capability, forward message as usual
            doForwardMessage(i,obj[i].fwdST.iocMessage);
            undefine obj[i].fwdST.iocMessage;
          end;

      else
        -- Save message to be forwarded
        obj[i].fwdST.Source      := inM.source;
        obj[i].fwdST.iocMessage  := inM;
        -- Create message for revoker
        undefine outM;
        outM.source       := i;
        outM.dest         := obj[i].fwdST.Revoker;
        outM.message      := M_Call;
        outM.onPath       := inM.onPath;
        outM.onPath[i]    := true;
        outM.isWrapped    := obj[i].isMembrane;
        obj[i].state      := O_BLOCK;
        addMessageToNet (outM);
      end;


    end;
  end;
end;

ruleset i: ObjectId do
  choose j: net do
    rule "Limited revoker/forarder object receives and forwards message"
      
      net[j].dest = i           &      -- Message for O_i
      obj[i].isForwarder        &      -- O_i is a forwarder
      obj[i].isRevoker          &      -- O_i is a revoker
      obj[i].isLimited          &      -- O_i is limited
      !obj[i].fwdST.isRevocable &      -- O_i is not revocable
      ((obj[i].state = O_IDLE & net[j].message != M_Return) |
       (obj[i].state = O_BLOCK& net[j].message = M_Return) )

    ==>

    var
      inM:  IOCMessage;
      outM: IOCMessage;

    begin
      inM := net[j];
      multisetremove (j,net);


      if inM.message = M_Revoke & inM.source = obj[i].fwdST.Revoker then
        updateTaintList(i,inM);
        --toggle flag:
        obj[i].revST.enabled := !obj[i].revST.enabled;
        obj[i].revST.changedAt := curTime;

        undefine outM;
        outM.source       := i;
        outM.dest         := inM.source;
        outM.message      := M_Return;
        outM.isWrapped    := inM.isWrapped | obj[i].isMembrane;

        resetPath(i,outM);
        outM.args.isRevoked := !obj[i].revST.enabled;
        addMessageToNet (outM); -- send Return([isRevoked])

      else 

        if obj[i].revST.enabled then
            -- Still have capability, forward message as usual
            doForwardMessage(i,inM);
        else
            updateTaintList(i,inM);
            -- Capability has been revoked, reply back to source with fail
            if obj[i].state = O_IDLE then
                obj[i].fwdST.Source := inM.source;
            end;
            doForwardFailMessage(i,inM);


        end;
      end;



    end;
  end;
end;

--------------------------------------------------------------------------------
-- startstate
--------------------------------------------------------------------------------

startstate
  undefine net;
  undefine obj;

  curTime := 0;

  for i : ObjectId do
    obj[i].state       := O_IDLE;       -- all objects are in idle state
    for j : ObjectId do
      obj[i].cList[j]  := CAP_NULL;     -- no capabily to any objects
      obj[i].taintList[j]  := false;    -- no communication has been tainted
    end;
    obj[i].cList[i]    := CAP_RAW;      -- always have capability to self
    obj[i].isForwarder := false;        -- none of the objects are forwarders
    obj[i].isMembrane  := false;        -- none of the objects are membranes
    obj[i].isRevoker   := false;        -- none of the objects are revokers 
    obj[i].isLimited   := false;        -- none of the objects are limited
  
    switch EXAMPLE_CASE
      case 1 : doExample1();
      case 2 : doExample2();
      case 3 : doExample3();
      case 4 : doExample4();

    end; 
  end;

end;

--------------------------------------------------------------------------------
-- | Invariants
--------------------------------------------------------------------------------

-- | Checking membrane pattern using taintList and onPath
--
-- Role of onPath:
-- If O_i sends a message to O_j through objects O_k, k \in K then
-- the message O_j receives has onPath[l] = true for l \in {i} U K
--
-- Role of taintList:
--  If O_j receives a raw message from O_i through objects O_k, k \in K then
--  the whole path is tainted, and so 
--  obj[j].taintList[obj[l]] = true for l \in {i} U K
--  
-- The membrane pattern is violated if a path including a membrane 
-- object to any other object is tainted. Specifically, for every membrane
-- object O_i if there is any object O_j such that O_j.taintList[O_i] is true,
-- the membrane pattern has been violated as O_j received a message through
-- O_i that was not wrapped.
--
invariant "(Forwarding) Membrane pattern holds"
  forall i : ObjectId do
    forall j : ObjectId do
      obj[i].isMembrane
      ->
      obj[j].taintList[i]=false -- path O_i to O_j is not tainted
    end
  end;

-- | Checking revocable (membrane) forwarder pattern holds
-- If "owner" sends a revoker object the Revoke() message then 
-- it should be that no message should be sent from the corresponding
-- forwarder unless the message is a Call to the revoker or is a fail Return
-- message.
--
invariant "Revocable Membrane Forwarder pattern holds"
  forall i : ObjectId do
    obj[i].isForwarder                          & -- O_i is forwarder
    obj[i].isMembrane                           & -- O_i is membrane
    !obj[i].isLimited                           & -- O_i is not limited
    obj[i].fwdST.isRevocable                    & -- O_i is revocable
    !obj[obj[i].fwdST.Revoker].revST.enabled      -- O_i is revoked
    ->
    multisetcount(j:net,
      net[j].source = i                   & -- source is O_i
      net[j].dest != obj[i].fwdST.Revoker & -- not to revoker
      ! (net[j].message = M_Return & net[j].args.isRevoked = true) -- not fail
      ) = 0
  end;

-- | Check same pattern as above with an additional restriction:
-- the revoker toggles the enable flag and then a message by the forwarder
-- is sent. This finds the same attack as Murray.
invariant "Revocable Membrane Forwarder pattern holds with time"
  forall i : ObjectId do
    obj[i].isForwarder                          & -- O_i is forwarder
    obj[i].isMembrane                           & -- O_i is membrane
    !obj[i].isLimited                           & -- O_i is not limited
    obj[i].fwdST.isRevocable                    & -- O_i is revocable
    !obj[obj[i].fwdST.Revoker].revST.enabled      -- O_i is revoked
    ->
    multisetcount(j:net,
      net[j].source = i                   & -- source is O_i
      net[j].dest != obj[i].fwdST.Revoker & -- not to revoker
      (obj[obj[i].fwdST.Revoker].revST.changedAt<net[j].tstamp) &
      ! (net[j].message = M_Return & net[j].args.isRevoked = true) -- not fail
      ) = 0
  end;


invariant "Limited Revocable/Forwarder pattern holds"
  forall i : ObjectId do
    obj[i].isForwarder              & -- O_i is forwarder
    obj[i].isRevoker                & -- O_i is forwarder
    obj[i].isLimited                & -- O_i is not limited
    !obj[i].fwdST.isRevocable       & -- O_i is revocable
    !obj[i].revST.enabled             -- O_i is revoked
    ->
    multisetcount(j:net,
      net[j].source = i                   & -- source is O_i
      net[j].dest != obj[i].fwdST.Revoker & -- not to revoker
      ! (net[j].message = M_Return & net[j].args.isRevoked = true) -- not fail
      ) = 0
  end;

-- No attack was found
-- 13720000 states explored in 430.29s, with 22573321 rules fired
-- and 4277327 states in the queue.

invariant "Limited Revocable/Forwarder pattern holds with time"
  forall i : ObjectId do
    obj[i].isForwarder              & -- O_i is forwarder
    obj[i].isRevoker                & -- O_i is forwarder
    obj[i].isLimited                & -- O_i is not limited
    !obj[i].fwdST.isRevocable       & -- O_i is revocable
    !obj[i].revST.enabled             -- O_i is revoked
    ->
    multisetcount(j:net,
      net[j].source = i                   & -- source is O_i
      net[j].dest != obj[i].fwdST.Revoker & -- not to revoker
      (obj[i].revST.changedAt<net[j].tstamp) &
      ! (net[j].message = M_Return & net[j].args.isRevoked = true) -- not fail
      ) = 0
  end;
