--------------------------------------------------------------------------------
--
--  Murphi Model of OCap patterns: Sealer/Unsealer
--  Deian Stefan
--
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- | Constants, types and variables
--------------------------------------------------------------------------------


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
const
  EXAMPLE_CASE: 1;
  NumObjects:   5;     -- number of objects
  NumSlots:     5;     -- number of slots
  NetworkSize:  2;     -- max. number of outstanding messages in network
  MAX_NR_MESSAGES: 100;
  FIXED: false;

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
type
  ObjectId:   1..NumObjects;
  SlotId:     1..NumSlots;
  TimeStamp:  0..MAX_NR_MESSAGES;

  
  Message : enum {   -- [] denotes optional, {} denotes implicit
    M_Unseal,        -- Unseal({return-capability}, capability)
    M_Call,          -- Call({return-capability})
    M_Return         -- Return([SlotContents])
  };

  SlotContent : enum {
    SLOT_NULL,  -- Slot is empty
    SLOT_CASH,  -- Slot has cash
    SLOT_FAKE   -- Slot has fake contents
  };

  SuperSlot : record
    content: SlotContent; -- Slot contents
    boxCap:  ObjectId;    -- Capability to box
  end;

  MessageArgs : record
    cap:       ObjectId; -- Capability
    slotRes:   SuperSlot; -- -- Returned slot content
    calledBox: ObjectId; -- 
  end;
  -- | Inter Object Communication message
  IOCMessage : record    
    source:    ObjectId;    -- Souce of message
    dest:      ObjectId;    -- Destination of message
    message:   Message;     -- Message
    args:      MessageArgs; -- Message arguments
    tstamp:    TimeStamp;  -- Time stamp
  end;

  -- | Object state
  ObjectState : enum {
    O_IDLE,     -- Object is in idle state
    O_BLOCK     -- Object waiting for receiver to return message
  };

  -- | Cabaility type
  CapType : enum {
    CAP_NULL,           -- No capability
    CAP_RAW             -- Unwrapped, i.e., raw capability,
  };



  -- | Objects.
  Object : record
    state:       ObjectState;                -- Current object state
    cList:       array[ObjectId] of CapType; -- List of capabilites

    -- Object is an unseal that clears slot and invokes the
    -- capability passed
    isUnsealer:  boolean;
      returnCap: ObjectId;
    -- Object is a box that puts boxContent in slot
    isBox:       boolean;
     boxContent: SuperSlot;
    --
    slot:        SlotId;
  end;


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var                                            -- state variables for
  net: multiset[NetworkSize] of IOCMessage;    --  network
  obj: array[ObjectId] of Object;              --  objects
  slots: array[SlotId] of SuperSlot;          --  slots
  --curTime: TimeStamp;                          --  time

--------------------------------------------------------------------------------
-- | Helper functions/procedured
--------------------------------------------------------------------------------

  procedure addMessageToNet ( var  m : IOCMessage );
  begin
    multisetadd(m,net);
  end;

--------------------------------------------------------------------------------
-- | Example start states
--------------------------------------------------------------------------------
  procedure doExample1();
  begin
  -- | Description:
  -- Attenuated forwarder pattern.
  -- O_4 is an attenuated forwarder to O_3.
  --                  O_2o-->[2: ]
  --                   ^ 
  --                   |
  --                   |
  --                   |
  --                   |
  --                   o 
  --                  O_1             O_5
  --                   o               o
  --                   |               |
  --                   |               |
  --                   |               |
  --                   |               |
  --                   v               v
  --                  O_3o-->[3: ]<--oO_4

    assert (NumObjects >= 5) "Example requires at least 5 objects";

    obj[1].cList[2] := CAP_RAW;       -- O_1 object has capability to O_2
    obj[1].cList[3] := CAP_RAW;       -- O_1 object has capability to O_3

    -- O_2 is a box with slot 2 to be filled with fake contents
    obj[2].isBox      := true;
    obj[2].boxContent.content := SLOT_FAKE;
    obj[2].slot       := 2;

    -- O_3 is an unsealer with access to slot 3
    obj[3].isUnsealer := true;
    obj[3].slot       := 3;

    -- O_4 is a box with slot 3 to be filled with real contents
    obj[4].isBox      := true;
    obj[4].boxContent.content := SLOT_CASH;
    obj[4].slot       := 3;

    obj[5].cList[4] := CAP_RAW;       -- O_5 object has capability to O_4


  end;


--------------------------------------------------------------------------------
-- | Rules
--------------------------------------------------------------------------------

-- | O_i sends Call({O_i}) to another object that is not an unsealer.
-- 
-- Once the message is sent, O_i blocks, waiting for return.
--
ruleset i: ObjectId do
  ruleset j: ObjectId do
    rule "Simple object sends Call message"
      
      i != j                      &     -- O_i is not O_j
      obj[i].state = O_IDLE       &     -- O_i in idle state
      !obj[i].isBox               &     -- O_i is not a box
      !obj[i].isUnsealer          &     -- O_i is not a unsealer
      !obj[j].isUnsealer          &     -- O_j is not a unsealer
      obj[i].cList[j] != CAP_NULL &     -- O_i has capability to O_j 

      multisetcount (l:net, true) < NetworkSize
      
    ==>

    var
      outM: IOCMessage;

    begin
      undefine outM;
      outM.source       := i;
      outM.dest         := j;
      outM.message      := M_Call;
      outM.args.cap     := i;   -- capability used to return

      obj[i].state      := O_BLOCK;
      addMessageToNet (outM);

    end;
  end;
end;

-- | Simple object O_i sends Unseal(O_k) to O_j, where O_k is a 
-- box object and O_j is an unsealer.
-- 
-- Once the message is sent, O_i blocks, waiting for return.
--
ruleset i: ObjectId do
  ruleset j: ObjectId do
    ruleset k: ObjectId do
      rule "Simple object sends Unseal message"
        
        i != j                      &     -- O_i is not O_j
        obj[i].state = O_IDLE       &     -- O_i in idle state
        !obj[i].isBox               &     -- O_i is not a box
        !obj[i].isUnsealer          &     -- O_i is not a unsealer
        obj[j].isUnsealer           &     -- O_j is a unsealer
        obj[k].isBox                &     -- O_k is a box
        obj[i].cList[j] != CAP_NULL &     -- O_i has capability to O_j 
        obj[i].cList[k] != CAP_NULL &     -- O_i has capability to O_k 

        multisetcount (l:net, true) < NetworkSize
        
      ==>

      var
        outM: IOCMessage;

      begin
        undefine outM;
        outM.source       := i;
        outM.dest         := j;
        outM.message      := M_Unseal;
        outM.args.cap     := k;

        obj[i].state      := O_BLOCK;
        addMessageToNet (outM);

      end;
    end;
  end;
end;

-- | Simple or box object O_i receives message.
-- If the message is a Call({O_j}) then O_i sends a Return() message to O_j.
--
-- If the message is a Return() then O_i goes back to idle.
--
-- If O_i is a box and message is not a Return(), then O_i fills its slot
-- with the content.
--
ruleset i: ObjectId do
  choose j: net do
    rule "Simple or box object receives message"
      
      net[j].dest = i            &      -- Message for O_i
      !obj[i].isUnsealer         &      -- O_i in not a unsealer
      net[j].message != M_Unseal &      -- Should not be Unseal(..)
      ((obj[i].state = O_IDLE & net[j].message != M_Return) |
       (obj[i].state = O_BLOCK& net[j].message = M_Return) ) 

    ==>

    var
      inM:  IOCMessage;
      outM: IOCMessage;

    begin
      inM := net[j];
      multisetremove (j,net);

      if inM.message = M_Return then
        obj[i].state      := O_IDLE;
      else

        undefine outM;
        outM.source       := i;
        outM.dest         := inM.source;
        outM.message      := M_Return;

        if obj[i].isBox then
          slots[obj[i].slot] := obj[i].boxContent;
          slots[obj[i].slot].boxCap := i;
        end;

        addMessageToNet (outM); -- send Return()
      end;

    end;
  end;
end;

ruleset i: ObjectId do
  choose j: net do
    rule "Unsealer receives/sends message"
      
      net[j].dest = i           &      -- Message for O_i
      obj[i].isUnsealer         &      -- O_i in a unsealer
      (net[j].message = M_Unseal | net[j].message = M_Return) &
      ((obj[i].state = O_IDLE & net[j].message != M_Return) |
       (obj[i].state = O_BLOCK& net[j].message = M_Return) ) 

    ==>

    var
      inM:  IOCMessage;
      outM: IOCMessage;

    begin
      inM := net[j];
      multisetremove (j,net);


      if inM.message = M_Return then
        undefine outM;
        outM.source         := i;
        outM.dest           := obj[i].returnCap;
        if !FIXED then
          outM.args.slotRes   := slots[obj[i].slot];
        else

          outM.args.slotRes.content := SLOT_NULL;
          if (!isundefined(slots[obj[i].slot].boxCap))
             & (slots[obj[i].slot].boxCap = inM.source) then
          outM.args.slotRes   := slots[obj[i].slot];
          end;
        end;
        outM.args.calledBox := inM.source;
        outM.message        := M_Return;

        obj[i].state        := O_IDLE;
        addMessageToNet (outM); -- send Return()
      else

        undefine outM;
        outM.source       := i;
        outM.dest         := inM.args.cap;
        outM.message      := M_Call;
        outM.args.cap     := i;   -- capability used to return

        obj[i].returnCap  := inM.source; -- save return cap:

        
        undefine slots[obj[i].slot];
        slots[obj[i].slot].content := SLOT_NULL;

        obj[i].state      := O_BLOCK;
        addMessageToNet (outM); -- send Return()
      end;

    end;
  end;
end;


--------------------------------------------------------------------------------
-- startstate
--------------------------------------------------------------------------------

startstate
  undefine net;
  undefine obj;
  undefine slots;

  --curTime := 0;

  for i : ObjectId do
    obj[i].state       := O_IDLE;       -- all objects are in idle state
    for j : ObjectId do
      obj[i].cList[j]  := CAP_NULL;     -- no capabily to any objects
    end;
    obj[i].cList[i]    := CAP_RAW;      -- always have capability to self
    obj[i].isUnsealer  := false;        -- none of the objects are unsealers
    obj[i].isBox       := false;        -- none of the objects are boxes
  
    switch EXAMPLE_CASE
      case 1 : doExample1();
    end; 
  end;

  for i : SlotId do
    slots[i].content := SLOT_NULL;
  end;

end;

--------------------------------------------------------------------------------
-- | Invariants
--------------------------------------------------------------------------------

invariant "Sealer/unsealer pattern holds"
  forall i : ObjectId do
      !obj[i].isUnsealer&
      !obj[i].isBox    
      ->
    multisetcount(n:net,
      net[n].dest = i                   & -- destination is O_i
      obj[net[n].source].isUnsealer     & -- source is an unsealer
      net[n].message = M_Return         & -- message is s Return
      net[n].args.slotRes.content = SLOT_CASH   & -- message argument is cash
      obj[net[n].args.calledBox].isBox  & -- from a box with a fake value
      obj[net[n].args.calledBox].boxContent.content = SLOT_FAKE
      ) = 0
  end;
